#ifndef __light_h__
#define __light_h__

#include "glut_import.h"

#include "list.h"
#include "handled.h"
#include "vect.h"

#define LF_COLOR                       0x0001
#define LF_GEOMETRY                    0x0002
#define LF_SPOT                        0x0003
#define LF_ATTENUATION                 0x0004
#define LF_ALL                         LF_COLOR | LF_GEOMETRY | LF_SPOT | LF_ATTENUATION

class Light: protected Handled<GLenum,0>{
  static List<Light *> Lights;
  static int MaxLights;
  unsigned int Modified;
  int Enabled;

  public:
    struct LightColor{
      Vect4<float> Ambient,Diffuse,Specular;
    };
    struct LightGeometry{
      int Directional;
      Vect4<float> Position;
      Vect3<float> Direction;
    };
    struct LightSpot{
      int Exponent,Cutoff;
    };
    struct LightAttenuation{
      float Constant,Linear,Quadratic;
    };

    static const LightColor white;
    static LightColor defColor(Vect4<float> color);

    static const LightGeometry upx;
    static const LightGeometry upy;
    static const LightGeometry upz;

    static const LightSpot omniSpot;

    static const LightAttenuation noAttenuation;
  protected:
    LightColor Color;
    LightGeometry Geometry;
    LightSpot Spot;
    LightAttenuation Attenuation;

    void createHandle(){
      int i;
      if (MaxLights==0)
        glGetIntegerv(GL_MAX_LIGHTS,&MaxLights);
      i=Lights.indexOf(NULL);
      if (i=-1){
        i=Lights.getCount();
        if (i>=MaxLights){
          printf("Lights number exceeded\n");
          exit(0xFFFFFFFF);
        }
        Lights.add(this);
      }else
        Lights[i]=this;
      Handle=GL_LIGHT0+i;
      glEnable(Handle);
      updateFlags();
    };

    void deleteHandle(){
      glDisable(Handle);
      Lights[Handle-GL_LIGHT0]=NULL;
    };

    void updateFlags(){
      if (handleAllocated()){
        if (Modified & LF_COLOR){
          glLightfv(Handle,GL_AMBIENT,&Color.Ambient.A);
          glLightfv(Handle,GL_DIFFUSE,&Color.Diffuse.A);
          glLightfv(Handle,GL_SPECULAR,&Color.Specular.A);
        }
        if (Modified & LF_GEOMETRY){
          if (Geometry.Directional){
            Vect4<float> v;
            v=-1.f * Geometry.Direction;
            v.D=0;
            glLightfv(Handle,GL_POSITION,&v.A);
            glLightfv(Handle,GL_SPOT_DIRECTION,&Geometry.Direction.A);
          }else{
            Geometry.Position.D=1;
            glLightfv(Handle,GL_POSITION,&Geometry.Position.A);
            glLightfv(Handle,GL_SPOT_DIRECTION,&Geometry.Direction.A);
          }
        }
        if (Modified & LF_SPOT){
          glLighti(Handle,GL_SPOT_EXPONENT,Spot.Exponent);
          glLighti(Handle,GL_SPOT_CUTOFF,Spot.Cutoff);
        }
        if (Modified & LF_ATTENUATION){
          glLightf(Handle,GL_CONSTANT_ATTENUATION,Attenuation.Constant);
          glLightf(Handle,GL_LINEAR_ATTENUATION,Attenuation.Linear);
          glLightf(Handle,GL_QUADRATIC_ATTENUATION,Attenuation.Quadratic);
        }
        Modified=0;
      }
    };

    void modified(unsigned int flags){
      Modified=Modified | flags;
      updateFlags();
    };

  public:
    Light(LightColor color=white,LightGeometry geometry=upz,LightSpot spot=omniSpot,LightAttenuation attenuation=noAttenuation){
      Color=color;
      Geometry=geometry;
      Spot=spot;
      Attenuation=attenuation;
      Enabled=0;
      Modified=LF_ALL;
    };

    virtual ~Light(){freeHandle();};

    void enable(){setEnabled(1);};
    void disable(){setEnabled(0);};
    void on(){setEnabled(1);};
    void off(){setEnabled(0);};
    int getEnabled(){return Enabled;};
    void setEnabled(int enabled){
      if (enabled ^ Enabled){
        if (enabled)
          handleNeeded();
        else
          freeHandle();
      }
    };

    void setColor(LightColor color){Color=color;modified(LF_COLOR);};
    void setColor(Vect3<float> ambient,Vect3<float> diffuse,Vect3<float> specular){
      Color.Ambient=ambient;Color.Diffuse=diffuse;Color.Specular=specular;modified(LF_COLOR);
    };
    void setAmbient(Vect4<float> ambient){Color.Ambient=ambient;modified(LF_COLOR);};
    void setDiffuse(Vect4<float> diffuse){Color.Diffuse=diffuse;modified(LF_COLOR);};
    void setSpecular(Vect4<float> specular){Color.Specular=specular;modified(LF_COLOR);};
    LightColor getColor(){return Color;};
    Vect4<float> getAmbient(){return Color.Ambient;};
    Vect4<float> getDiffuse(){return Color.Diffuse;};
    Vect4<float> getSpecular(){return Color.Specular;};

    void setGeometry(LightGeometry geometry){Geometry=geometry;modified(LF_GEOMETRY);};
    void setGeometry(int directional,Vect3<float> position,Vect3<float> direction){
      Geometry.Directional=directional;Geometry.Position=position;Geometry.Direction=direction;modified(LF_GEOMETRY);
    };
    void setDirectionalGeometry(Vect3<float> direction){
      Geometry.Directional=1;Geometry.Direction=direction;modified(LF_GEOMETRY);
    };
    void setPositionalGeometry(Vect3<float> position,Vect3<float> direction){
      Geometry.Directional=0;Geometry.Position=vect(position.A,position.B,position.C);Geometry.Direction=direction;modified(LF_GEOMETRY);
    };
    void setDirectional(int directional){Geometry.Directional=directional;modified(LF_GEOMETRY);};
    void setPosition(Vect3<float> position){Geometry.Position=position;modified(LF_GEOMETRY);};
    void setDirection(Vect3<float> direction){Geometry.Direction=direction;modified(LF_GEOMETRY);};
    LightGeometry getGeometry(){return Geometry;};
    int getDirectional(){return Geometry.Directional;};
    Vect3<float> getPosition(){return Geometry.Position;};
    Vect3<float> getDirection(){return Geometry.Direction;};

    void setSpot(LightSpot spot){Spot=spot;modified(LF_SPOT);};
    void setSpot(int exponent,int cutoff){Spot.Exponent=exponent;Spot.Cutoff=cutoff;modified(LF_SPOT);};
    void setSpotExponent(int exponent){Spot.Exponent=exponent;modified(LF_SPOT);};
    void setSpotCutoff(int cutoff){Spot.Cutoff=cutoff;modified(LF_SPOT);};
    LightSpot getSpot(){return Spot;};
    int getSpotExponent(){return Spot.Exponent;};
    int getSpotCutoff(){return Spot.Cutoff;};

    void setAttenuation(LightAttenuation attenuation){Attenuation=attenuation;modified(LF_ATTENUATION);};
    void setAttenuation(float constant,float linear,float quadratic){
      Attenuation.Constant=constant;Attenuation.Linear=linear;Attenuation.Quadratic=quadratic;modified(LF_ATTENUATION);
    };
    void setConstantAttenuation(float attenuation){Attenuation.Constant=attenuation;modified(LF_ATTENUATION);};
    void setLinearAttenuation(float attenuation){Attenuation.Linear=attenuation;modified(LF_ATTENUATION);};
    void setQuadraticAttenuation(float attenuation){Attenuation.Quadratic=attenuation;modified(LF_ATTENUATION);};
    LightAttenuation getAttenuation(){return Attenuation;};
    float getConstantAttenuation(){return Attenuation.Constant;};
    float getLinearAttenuation(){return Attenuation.Linear;};
    float getQuadraticAttenuation(){return Attenuation.Quadratic;};
};

List<Light *> Light::Lights;
int Light::MaxLights=0;

const Light::LightColor Light::white={{0,0,0,1},{1,1,1,1},{1,1,1,1}};

Light::LightColor Light::defColor(Vect4<float> color){
  Light::LightColor c;
  c.Ambient=vect(0,0,0,1);
  c.Diffuse=color;
  c.Specular=color;
  return c;
};

const Light::LightGeometry Light::upx={1,{0,0,0},{-1,0,0}};
const Light::LightGeometry Light::upy={1,{0,0,0},{0,-1,0}};
const Light::LightGeometry Light::upz={1,{0,0,0},{0,0,-1}};

const Light::LightSpot Light::omniSpot={0,180};

const Light::LightAttenuation Light::noAttenuation={1,0,0};

#endif
